# Copyright 2015 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
''' Provide atomic file support - a subset of `fatomic` included here
(https://github.com/abarnert/fatomic) as upstream is not in pypi
'''
import contextlib
import errno
import os
import shutil
import sys
import tempfile
import types

# If we're on 3.3+, just use os.replace; if we're on POSIX, rename
# and replace do the same thing.
try:
    _replace_file = os.replace
except AttributeError:
    if sys.platform != 'win32':
        _replace_file = os.rename
    else:
        # This requires PyWin32 if you're on Windows. If that's not
        # accepted, you can write a ctypes solution, but then you'll
        # have to handle unicode-vs.-bytes strings and creating an
        # OSError from GetLastError and so on yourself, which I don't
        # feel like doing. (I'll accept a pull request from anyone
        # else who does...)
        import win32api
        import win32con

        def _replace_file(src, dst):
            win32api.MoveFileEx(src, dst, win32con.MOVEFILE_REPLACE_EXISTING)


@contextlib.contextmanager
def atomic_open(filename, mode, *args, **kwargs):
    if mode[0] not in 'wxa' or len(mode) > 1 and mode[1] == '+':
        raise ValueError("invalid mode: '{}'".format(mode))
    f = tempfile.NamedTemporaryFile(mode=mode,
                                    prefix=os.path.basename(filename),
                                    dir=os.path.dirname(filename),
                                    suffix='.tmp',
                                    delete=False)
    # track: explicitly discarded, normal/abnormal completion
    _discard = [None]
    try:
        if mode[0] == 'a':
            try:
                with open(filename, 'r'+mode[1:], *args, **kwargs) as fin:
                    shutil.copyfileobj(fin, f)
            except (OSError, IOError) as e:
                if e.errno == errno.ENOENT:
                    pass

        def discard(self, _discard=_discard):
            # explicit discard
            _discard[0] = True
        f.discard = types.MethodType(discard, f)
        yield f
        # normal completion
        if not _discard[0]:
            _discard[0] = False
    # block and force discarding
    finally:
        f.close()
        # if we didn't complete or were aborted, delete
        if _discard[0] is None or _discard[0]:
            os.unlink(f.name)
        else:
            _replace_file(f.name, filename)



# Copyright 2015 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from .utils import read_planet_json

ENV_KEY = 'PL_API_KEY'


class APIKey(object):
    def __init__(self, value):
        self.value = value


def find_api_key():
    api_key = os.getenv(ENV_KEY)
    if api_key is None:
        contents = read_planet_json()
        api_key = contents.get('key', None)
    return api_key



# Copyright 2017 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import logging
import sys

from planet import api
from planet.api.__version__ import __version__
from planet.api.utils import write_planet_json
from .util import call_and_wrap

client_params = {}


def clientv1():
    return api.ClientV1(**client_params)


def configure_logging(verbosity):
    '''configure logging via verbosity level of between 0 and 2 corresponding
    to log levels warning, info and debug respectfully.'''
    log_level = max(logging.DEBUG, logging.WARNING - logging.DEBUG*verbosity)
    logging.basicConfig(
        stream=sys.stderr, level=log_level,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )

    urllib3_logger = logging.getLogger(
        'requests.packages.urllib3')
    urllib3_logger.setLevel(log_level)

    # if debug level set then its nice to see the headers of the request
    if log_level == logging.DEBUG:
        try:
            import http.client as http_client
        except ImportError:
            # Python 2
            import httplib as http_client
        http_client.HTTPConnection.debuglevel = 1


@click.group()
@click.pass_context
@click.option('-w', '--workers', default=4,
              help=('The number of concurrent downloads when requesting '
                    'multiple scenes. - Default 4'))
@click.option('-v', '--verbose', count=True, help='Specify verbosity')
@click.option('-k', '--api-key',
              help='Valid API key - or via ENV variable %s' % api.auth.ENV_KEY)
@click.option('-u', '--base-url', envvar='PL_API_BASE_URL',
              help='Change the base Planet API URL or ENV PL_API_BASE_URL'
                   ' - Default https://api.planet.com/')
@click.version_option(version=__version__, message='%(version)s')
def cli(context, verbose, api_key, base_url, workers):
    '''Planet API Client'''

    configure_logging(verbose)

    client_params.clear()
    client_params['api_key'] = api_key
    client_params['workers'] = workers
    if base_url:
        client_params['base_url'] = base_url


@cli.command('help')
@click.argument("command", default="")
@click.pass_context
def help(context, command):
    '''Get command help'''
    if command:
        cmd = cli.commands.get(command, None)
        if cmd:
            context.info_name = command
            click.echo(cmd.get_help(context))
        else:
            raise click.ClickException('no command: %s' % command)
    else:
        click.echo(cli.get_help(context))


@cli.command('init')
@click.option('--email', default=None, prompt=True, help=(
    'The email address associated with your Planet credentials.'
))
@click.option('--password', default=None, prompt=True, hide_input=True, help=(
    'Account password. Will not be saved.'
))
def init(email, password):
    '''Login using email/password'''
    response = call_and_wrap(clientv1().login, email, password)
    write_planet_json({'key': response['api_key']})
    click.echo('initialized')



# Copyright 2017 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import threading
import time
from .utils import write_to_file
from planet.api.exceptions import (RequestCancelled, NoPermission)
try:
    import Queue as queue
except ImportError:
    # renamed in 3
    import queue


def _by_status(assets, types, status):
    return [assets[t] for t in types if
            t in assets and assets[t]['status'] == status]


def _all_status(assets, types, statuses):
    return all([assets[t]['status'] in statuses for t in types if t in assets])


_logger = logging.getLogger(__name__)
_debug = _logger.debug
_info = _logger.info


class _Stage(object):
    '''A _Stage performs some sequence in an activate/poll/download cycle.

    The stage will first attempt to process any existing tasks and if there
    is capacity, poll for another from the iterator-like source (a _Stage is
    one, itself). If the response is False, this implies the source is
    exhausted or cancelled. If the response is None, there are no pending
    tasks from the source.

    When a _Stage has completed a task, it places the output in _results so
    they are available. If the task needs repeating, it is placed back in the
    task queue.

    A _Stage may perform a maximum number of do-per-seconds to avoid throttling
    The current approach is conservative and does not issue RPS but instead
    throttles if a request complets in under the allotted time.

    The Downloader uses the following stages:

    search-iter -> activation -> polling -> download

    Downloader synchronously drains the download stage allowing for completion
    and cancellation of any in-flight backgrounded requests.

    Where search-iter is an iterator of 'item' features from the API.

    This approach allows for back-pressure from constrained stages as well as
    re-queueing tasks without being deadlock prone (e.g. pull vs. push) and
    simplifies cancellation of the entire pipeline.
    '''
    def __init__(self, source, size=0, max_dps=0):
        self._source = source
        self._running = True
        self._cancelled = False
        self._size = size
        self._tasks = []
        # track the current task
        self._doing = None
        self._results = queue.Queue()
        self._min_sleep = 1. / max_dps if max_dps else 0
        self._cond = threading.Condition()

    def work(self):
        return len(self._tasks) + (1 if self._doing else 0)

    def start(self):
        threading.Thread(target=self._run).start()

    def next(self):
        try:
            return self._results.get()
        except queue.Empty:
            if not self._alive():
                return False

    def _i(self, msg, *args):
        _info(type(self).__name__ + ' : ' + msg, *args)

    def _d(self, msg, *args):
        _debug(type(self).__name__ + ' : ' + msg, *args)

    def _cancel(self, result):
        pass

    def cancel(self):
        # this makes us not alive
        self._cancelled = True
        self._running = False
        self._tasks = []
        self._doing = None
        # drain any results and cancel them
        while not self._results.empty():
            self._results.get()
        self._results.put(False)
        # notify any sleepers
        try:
            self._cond.acquire()
            self._cond.notify_all()
        finally:
            self._cond.release()

    def _task(self, t):
        return t

    def _alive(self):
        # alive means upstream source has pending stuff or neither upstream
        # or this stage have been cancelled or we have some pending tasks
        # in the case upstream is done, but we're not
        return not self._cancelled and (
            self._running or len(self._tasks) or self._doing
        )

    def _capacity(self):
        return len(self._tasks) < self._size

    def _get_tasks(self):
        while self._capacity() and self._running:
            try:
                # @todo minor refactor
                # bleh, next wants an iterator, not just a __next__
                if hasattr(self._source, 'next'):
                    n = self._source.next()
                else:
                    n = next(self._source)
            except StopIteration:
                n = False
            # upstream is done if False
            self._running = n is not False
            if n:
                n = self._task(n)
                n and self._tasks.insert(0, n)
            else:
                break

    def _process_task(self):
        if self._tasks:
            self._doing = self._tasks.pop(0)
            try:
                self._do(self._doing)
            except Exception:
                # @todo should cancel the entire process?
                self._running = False
                logging.exception('unexpected error in %s', self)
                return
            self._doing = None

    def _run(self):
        while self._alive():
            self._get_tasks()
            t = time.time()
            self._process_task()
            # note - this is conservative compared to timer invocation.
            # allow _at most_ 1 'do' per min_sleep
            wait = self._min_sleep - (time.time() - t)
            if wait > 0 and not self._cancelled:
                self._d('sleep for %.2f', wait)
                # waiting on the condition allows interrupting sleep
                self._cond.acquire()
                self._cond.wait(wait)
                self._cond.release()

        # sentinel value to indicate we're done
        self._results.put(False)


class _AStage(_Stage):
    def __init__(self, source, client, asset_types):
        _Stage.__init__(self, source, 100, max_dps=5)
        self._client = client
        self._asset_types = asset_types

    def _do(self, item):
        assets = self._client.get_assets(item).get()
        if not any([t in assets for t in self._asset_types]):
            _info('no desired assets in item, skipping')
            return
        inactive = _by_status(assets, self._asset_types, 'inactive')
        if inactive:
            # still need activation, try the first inactive
            self._client.activate(inactive[0])
            self._tasks.append(item)
            return

        if _all_status(assets, self._asset_types, ['activating', 'active']):
            self._results.put((item, assets))
        else:
            # hmmm
            status = [assets[t]['status'] for t in self._asset_types]
            raise Exception('unexpected state %s' % status)


class _PStage(_Stage):
    _min_poll_interval = 5

    def __init__(self, source, client, asset_types):
        _Stage.__init__(self, source, 100, max_dps=2)
        self._client = client
        self._asset_types = asset_types

    def _task(self, t):
        item, assets = t
        now = time.time()
        return item, assets, now, now

    def _do(self, task):
        item, assets, start, last = task
        now = time.time()
        # don't poll until min interval elapsed
        if now - last > self._min_poll_interval:
            assets = self._client.get_assets(item).get()
            last = now
        if _all_status(assets, self._asset_types, ['active']):
            _debug('activation took %d', time.time() - start)
            self._results.put((item, assets))
        else:
            self._tasks.append((item, assets, start, last))


class _DStage(_Stage):
    def __init__(self, source, client, asset_types, dest):
        # @todo max pool should reflect client workers
        _Stage.__init__(self, source, 4, max_dps=2)
        self._client = client
        self._asset_types = asset_types
        self._dest = dest
        self._write_lock = threading.Lock()
        self._written = 0
        self._downloads = 0

    def _task(self, t):
        item, assets = t
        for at in self._asset_types:
            self._tasks.append((item, assets[at]))

    def cancel(self):
        while not self._results.empty():
            try:
                r = self._results.get(block=False)
                if r:
                    item, asset, dl = r
                    dl.cancel()
            except queue.Empty:
                pass
        _Stage.cancel(self)

    def _write_tracker(self, item, asset):
        def _tracker(**kw):
            if 'skip' in kw:
                self._i('skipping download of %s, already exists',
                        kw['skip'].name)
            elif 'wrote' in kw:
                with self._write_lock:
                    self._written += kw['wrote']
        return _tracker

    def _get_writer(self, item, asset):
        return

    def _do(self, task):
        item, asset = task
        writer = write_to_file(
            self._dest, self._write_tracker(item, asset), overwrite=False)
        self._downloads += 1
        self._results.put((item, asset,
                           self._client.download(asset, writer)))


class Downloader(object):
    '''A Downloader manages activation and download of Item Assets from the
    Data API. A Downloader should only be processing one request to either
    `activate` or `download` at a time. These functions are synchronous and
    will return on completion. Completion of activation or download events
    can be tracked by changing the `on_complete` method of the Downloader while
    the `stats` function allows for polling of internal state.
    '''

    def shutdown(self):
        '''Halt execution.'''
        raise NotImplementedError()

    def stats(self):
        '''Retrieve internal state of the Downloader.

        Returns a dict of state:

        - paging: `bool` indicating that search results are being processed
        - activating: `int` number of items in the inactive or activating state
        - downloading: `int` number of items actively downloading
        - downloaded: `string` representation of MB transferred
        - complete: `int` number of completed downloads
        - pending: `int` number of items awaiting download
        '''
        raise NotImplementedError()

    def activate(self, items, asset_types):
        '''Request activation of specified asset_types for the sequence of
        items.

        :param items: a sequence of Item representations.
        :param asset_types list: list of asset-type (str)
        '''
        raise NotImplementedError()

    def download(self, items, asset_types, dest):
        '''Request activation and download of specified asset_types for the
        sequence of items.

        :param items: a sequence of Item representations.
        :param asset_types list: list of asset-type (str)
        :param dest str: Download destination directory, must exist.
        '''
        raise NotImplementedError()

    def on_complete(self, item, asset, path=None):
        '''Notification of processing an item's asset, invoked on completion of
        `activate` or `download`.

        :param item: The API representation of the item
        :param asset: The API representation of the asset
        :param path: If downloaded, the location of the downloaded asset,
                     otherwise None
        '''
        pass


class _Downloader(Downloader):
    def __init__(self, client, **opts):
        self._client = client
        self._opts = opts
        self._stages = []
        self._completed = 0
        self._waiting = None

    def activate(self, items, asset_types):
        return self._run(items, asset_types)

    def download(self, items, asset_types, dest):
        return self._run(items, asset_types, dest)

    def _init(self, items, asset_types, dest):
        client = self._client
        astage = _AStage(items, client, asset_types)
        pstage = _PStage(astage, client, asset_types)
        self._stages = [
            astage,
            pstage
        ]
        if dest:
            dstage = _DStage(pstage, client, asset_types, dest)
            self._stages.append(dstage)
            self._dest = dest

        # sneaky little hack to allow tests to inject options
        self._apply_opts(vars())
        self._completed = 0

    def _run(self, items, asset_types, dest=None):
        if self._stages:
            raise Exception('already running')

        self._init(items, asset_types, dest)

        [s.start() for s in self._stages]

        last = self._stages[-1]
        while self._stages:
            try:
                n = last.next()
                if n is False:
                    break
                # this represents an activation completion, report
                # each requested item/asset combo
                # @todo hacky lack of internal structure in results
                if len(n) == 2:
                    item, assets = n
                    for a in asset_types:
                        self.on_complete(item, assets[a])
                # otherwise it is a download
                else:
                    item, asset, self._waiting = n
                    try:
                        body = self._waiting.wait()
                        self._waiting = None
                        dl = os.path.join(self._dest, body.name)
                        self.on_complete(item, asset, dl)
                    except RequestCancelled:
                        pass
                self._completed += 1
            except StopIteration:
                break
        stats = self.stats()
        self._stages = []
        return stats

    def _apply_opts(self, to):
        opts = self._opts
        opt = opts.pop('no_sleep', False)
        if opt:
            for s in self._stages:
                s._min_sleep = 0
        for k in opts:
            v, a = k.split('_', 1)
            t = to.get(v, None)
            if t is None:
                raise Exception('option not supported %s', k)
            else:
                setattr(t, a, opts[k])

    def stats(self):
        stats = {
            'paging': False,
            'activating': 0,
            'pending': 0,
            'complete': 0,
        }
        if len(self._stages) == 3:
            stats['downloading'] = 0
            stats['downloaded'] = '0.0MB'
        if not self._stages:
            return stats

        astage, pstage = self._stages[:2]
        dstage = None if len(self._stages) == 2 else self._stages[2]
        if dstage is not None:
            mb_written = '%.2fMB' % (dstage._written / 1.0e6)
            stats['downloading'] = dstage._downloads - self._completed
            stats['downloaded'] = mb_written
        stats['paging'] = astage._running
        stats['activating'] = astage.work() + pstage.work()
        stats['pending'] = (dstage.work() if dstage else 0)
        stats['complete'] = self._completed
        return stats

    def shutdown(self):
        for s in self._stages:
            s.cancel()
        self._waiting and self._waiting.cancel()
        self._stages = []
        self._client.shutdown()


class _MosaicDownloadStage(_DStage):
    def _task(self, t):
        return t

    def _do(self, task):
        func = self._write_tracker(task, None)
        writer = write_to_file(self._dest, func, overwrite=False)
        try:
            self._results.put((task, {'type': 'quad'},
                               self._client.download_quad(task, writer)))
            self._downloads += 1
        except NoPermission:
            _info('No download permisson for %s, skipping', task['id'])


class _MosaicDownloader(_Downloader):
    def activate(self, items, asset_types):
        pass

    def _init(self, items, asset_types, dest):
        client = self._client
        dstage = _MosaicDownloadStage(items, client, asset_types, dest)
        self._dest = dest
        self._stages.append(dstage)
        self._apply_opts(vars())
        self._completed = 0

    def stats(self):
        stats = {
            'paging': False,
            'activating': 0,
            'pending': 0,
            'complete': 0,
            'downloading': 0,
            'downloaded': '0.0MB',
        }
        if not self._stages:
            return stats

        dstage = self._stages[0]
        mb_written = '%.2fMB' % (dstage._written / float(1024**2))
        stats['downloading'] = dstage._downloads - self._completed
        stats['downloaded'] = mb_written
        stats['pending'] = dstage.work()
        stats['complete'] = self._completed
        return stats


class _OrderDownloadStage(_DStage):
    def _task(self, t):
        return t

    def _do(self, task):
        func = self._write_tracker(task, None)
        writer = write_to_file(self._dest, func, overwrite=False)
        self._downloads += 1
        self._results.put((task,
                           self._client.download_location(task, writer)))


class _OrderDownloader(_Downloader):
    def activate(self, items, asset_types):
        pass

    def _init(self, items, asset_types, dest):
        client = self._client
        dstage = _OrderDownloadStage(items, client, asset_types, dest)
        self._dest = dest
        self._stages.append(dstage)
        self._apply_opts(vars())
        self._completed = 0

    def stats(self):
        stats = {
            'paging': False,
            'activating': 0,
            'pending': 0,
            'complete': 0,
            'downloading': 0,
            'downloaded': '0.0MB',
        }
        if not self._stages:
            return stats

        dstage = self._stages[0]
        mb_written = '%.2fMB' % (dstage._written / float(1024**2))
        stats['downloading'] = dstage._downloads - self._completed
        stats['downloaded'] = mb_written
        stats['pending'] = dstage.work()
        stats['complete'] = self._completed
        return stats


def create(client, mosaic=False, order=False, **kw):
    '''Create a Downloader with the provided client.

    :param mosaic bool: If True, the Downloader will fetch mosaic quads.
    :returns: :py:Class:`planet.api.downloader.Downloader`
    '''
    if mosaic:
        return _MosaicDownloader(client, **kw)
    elif order:
        return _OrderDownloader(client, **kw)
    else:
        return _Downloader(client, **kw)



# Copyright 2015 Planet Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class APIException(Exception):
    '''General unexpected response'''
    pass


class BadQuery(APIException):
    '''Invalid inputs, HTTP 400'''
    pass


class InvalidAPIKey(APIException):
    '''Invalid key, HTTP 401'''
    pass


class NoPermission(APIException):
    '''Insufficient permissions, HTTP 403'''
    pass


class MissingResource(APIException):
    '''Request for non existing resource, HTTP 404'''
    pass


class TooManyRequests(APIException):
    '''Too many requests, HTTP 429'''
    pass


class OverQuota(APIException):
    '''Quota exceeded, HTTP 429'''
    pass


class ServerError(APIException):
    '''Unexpected internal server error, HTTP 500'''
    pass


class InvalidIdentity(APIException):
    '''Raised when logging in with invalid credentials'''
    pass


class RequestCancelled(Exception):
    '''Internal exception when a request is cancelled'''
    pass



import requests
from planet.api.dispatch import _get_user_agent
from planet.api.auth import find_api_key

ITEM_TYPE_URL = 'https://api.planet.com/data/v1/item-types/'
ASSET_TYPE_URL = 'https://api.planet.com/data/v1/asset-types/'

_item_types = None
_asset_types = None
_bundles = None

# Default values here are used as a fallback
# In case the API fails to respond or takes too long.
DEFAULT_ITEM_TYPES = [
    'Landsat8L1G',
    'MOD09GA',
    'MOD09GQ',
    'MYD09GA',
    'MYD09GQ',
    'PSOrthoTile',
    'PSScene',
    'PSScene3Band',
    'PSScene4Band',
    'REOrthoTile',
    'REScene',
    'Sentinel1',
    'Sentinel2L1C',
    'SkySatCollect',
    'SkySatScene',
    'SkySatVideo'
]

DEFAULT_ASSET_TYPES = [
    'ortho_analytic_4b', 'analytic_5b', 'analytic_5b_xml', 'ortho_analytic_8b',
    'ortho_analytic_8b_xml','ortho_analytic_8b_sr', 'analytic_b1', 'analytic_b10', 'analytic_b11',
    'analytic_b12', 'analytic_b2', 'analytic_b3', 'analytic_b4', 'analytic_b5',
    'analytic_b6', 'analytic_b7', 'analytic_b8', 'analytic_b8a', 'analytic_b9',
    'analytic_bqa', 'analytic_dn', 'analytic_dn_xml', 'analytic_gflags',
    'analytic_granule_pnt', 'analytic_iobs_res', 'analytic_ms',
    'analytic_num_observations', 'analytic_num_observations_1km',
    'analytic_num_observations_500m', 'analytic_obscov',
    'analytic_obscov_500m', 'analytic_orbit_pnt', 'analytic_q_scan',
    'analytic_qc_250m', 'analytic_qc_500m', 'analytic_range',
    'analytic_sensor_azimuth', 'analytic_sensor_zenith',
    'analytic_solar_azimuth', 'analytic_solar_zenith', 'ortho_analytic_4b_sr',
    'analytic_state_1km', 'analytic_sur_refl_b01', 'analytic_sur_refl_b02',
    'analytic_sur_refl_b03', 'analytic_sur_refl_b04', 'analytic_sur_refl_b05',
    'analytic_sur_refl_b06', 'analytic_sur_refl_b07', 'ortho_analytic_4b_xml',
    'basic_analytic_4b', 'basic_analytic_8b', 'basic_analytic_8b_xml',
    'basic_analytic_b1', 'basic_analytic_b1_nitf', 'basic_analytic_b2',
    'basic_analytic_b2_nitf', 'basic_analytic_b3', 'basic_analytic_b3_nitf',
    'basic_analytic_b4', 'basic_analytic_b4_nitf', 'basic_analytic_b5',
    'basic_analytic_b5_nitf', 'basic_analytic_dn', 'basic_analytic_dn_nitf',
    'basic_analytic_dn_rpc', 'basic_analytic_dn_rpc_nitf',
    'basic_analytic_dn_xml', 'basic_analytic_dn_xml_nitf',
    'basic_analytic_nitf', 'basic_analytic_4b_rpc', 'basic_analytic_rpc_nitf',
    'basic_analytic_sci', 'basic_analytic_udm2', 'basic_analytic_4b_xml',
    'basic_analytic_xml_nitf', 'basic_anlytic_udm', 'basic_l1a_all_frames',
    'basic_l1a_panchromatic_dn', 'basic_l1a_panchromatic_dn_rpc',
    'basic_panchromatic', 'basic_panchromatic_dn', 'basic_panchromatic_dn_rpc',
    'basic_panchromatic_rpc', 'basic_panchromatic_udm2', 'basic_udm',
    'basic_udm2', 'browse', 'metadata_aux', 'metadata_txt', 'ortho_analytic',
    'ortho_analytic_dn', 'ortho_analytic_hh', 'ortho_analytic_hv',
    'ortho_analytic_sr', 'ortho_analytic_udm', 'ortho_analytic_udm2',
    'ortho_analytic_vh', 'ortho_analytic_vv', 'ortho_panchromatic',
    'ortho_panchromatic_dn', 'ortho_panchromatic_udm',
    'ortho_panchromatic_udm2', 'ortho_pansharpened', 'ortho_pansharpened_udm',
    'ortho_pansharpened_udm2', 'ortho_visual', 'ps3b_analytic',
    'ps3b_analytic_dn', 'ps3b_analytic_dn_xml', 'ps3b_analytic_xml',
    'ps3b_basic_analytic', 'ps3b_basic_analytic_dn',
    'ps3b_basic_analytic_dn_rpc', 'ps3b_basic_analytic_dn_xml',
    'ps3b_basic_analytic_rpc', 'ps3b_basic_analytic_xml', 'ps3b_udm',
    'ps3b_visual_xml', 'ps4b_analytic_dn', 'ps4b_analytic_dn_xml',
    'ps4b_basic_analytic_dn', 'ps4b_basic_analytic_dn_nitf',
    'ps4b_basic_analytic_dn_rpc', 'ps4b_basic_analytic_dn_rpc_nitf',
    'ps4b_basic_analytic_dn_xml', 'ps4b_basic_analytic_dn_xml_nitf',
    'ps4b_basic_analytic_nitf', 'ps4b_basic_analytic_rpc_nitf',
    'ps4b_basic_analytic_xml_nitf', 'ps4b_udm', 'udm', 'ortho_udm2', 'video_file',
    'video_frames', 'video_metadata', 'ortho_visual', 'visual_xml'
]


DEFAULT_BUNDLES = [u'all', u'all_udm2', u'analytic', u'analytic_sr',
                   u'analytic_sr_udm2', u'analytic_udm2', u'basic_analytic',
                   u'basic_analytic_nitf', u'basic_analytic_nitf_udm2',
                   u'basic_analytic_udm2', u'basic_panchromatic',
                   u'basic_panchromatic_dn', u'basic_uncalibrated_dn',
                   u'basic_uncalibrated_dn_nitf',
                   u'basic_uncalibrated_dn_nitf_udm2',
                   u'basic_uncalibrated_dn_udm2', u'panchromatic',
                   u'panchromatic_dn', u'panchromatic_dn_udm2',
                   u'pansharpened', u'pansharpened_udm2', u'skysatvideo',
                   u'uncalibrated_dn', u'uncalibrated_dn_udm2', u'visual']


def _get_json_or_raise(url, timeout=11):
    api_key = find_api_key()
    headers = {'User-Agent': _get_user_agent(),
               'Authorization': 'api-key %s' % api_key}
    resp = requests.get(url, timeout=timeout, headers=headers)
    resp.raise_for_status()
    return resp.json()


def get_item_types():
    global _item_types
    if _item_types is None:
        _item_types = DEFAULT_ITEM_TYPES
        data = _get_json_or_raise(ITEM_TYPE_URL)
        _item_types = [it['id'] for it in data['item_types']]
    return _item_types


def get_asset_types():
    global _asset_types
    if _asset_types is None:
        _asset_types = DEFAULT_ASSET_TYPES
        data = _get_json_or_raise(ASSET_TYPE_URL)
        _asset_types = list(sorted([a['id'] for a in data['asset_types']]))
    return _asset_types


def get_bundles():
    global _bundles
    if _bundles is None:
        _bundles = DEFAULT_BUNDLES
        # TODO if/when bundles defs are served by API we can grab them here
    return _bundles
